from glotter.source import get_sources, filter_sources
from glotter.settings import Settings


def run(args):
    sources_by_type = filter_sources(args, get_sources(Settings().source_root))
    for project_type, sources in sources_by_type.items():
        params = _prompt_params(project_type)
        for source in sources:
            _build_and_run(source, params)


def _prompt_params(project_type):
    if not Settings().projects[project_type].requires_parameters:
        return ""
    return input(f'input parameters for "{project_type}": ')


def _build_and_run(source, params):
    print()
    print(f'Running "{source.name}{source.extension}"...')
    try:
        source.build()
        print(source.run(params))
    finally:
        source.cleanup()



import sys


def quote(value: str) -> str:
    """
    Enclose a value in quotes

    :param value: Value to enclosed
    :return: Quoted value

    """

    if '"' in value:
        if "'" in value:
            quote_chars = '"""'
            if value.startswith('"'):
                value = f"\\{value}"
        else:
            quote_chars = "'"
    else:
        quote_chars = '"'

    return f"{quote_chars}{value}{quote_chars}"


def indent(value: str, num_spaces: int) -> str:
    """
    Indent each line of a string by a specified number of spaces

    :param value: String to indent
    :param num_spaces: Number of spaces to indent
    :return: Indented string
    """

    spaces = " " * num_spaces
    return "".join(f"{spaces}{line}" for line in value.splitlines(keepends=True))


def error_and_exit(msg):
    print(msg)
    sys.exit(1)



import os
import shlex

from glotter.settings import Settings
from glotter.utils import quote


def generate_test_docs(doc_dir, repo_name, repo_url):
    """
    Generate test documentation for all projects

    :param doc_dir: Documentation directory
    :param repo_name: Repository name
    :param repo_url: Repository URL
    """

    settings = Settings()
    for project in settings.projects.values():
        test_doc_generator = TestDocGenerator(project)
        doc = test_doc_generator.generate_test_doc(repo_name, repo_url)
        if doc:
            project_dir = os.path.join(doc_dir, "-".join(project.words))
            os.makedirs(project_dir)
            project_doc_path = os.path.join(project_dir, "testing.md")
            with open(os.path.join(project_doc_path), "w", encoding="utf-8") as f:
                f.write(doc)


class TestDocGenerator:
    __test__ = False  # Indicate this is not a test

    def __init__(self, project):
        self.project = project
        self.project_title = " ".join(self.project.words).title()

    def generate_test_doc(self, repo_name, repo_url):
        if not self.project.tests:
            return ""

        doc = self._get_test_intro(repo_name, repo_url)
        if self.project.requires_parameters:
            for test_obj in self.project.tests.values():
                test_doc_section_generator = TestDocSectionGenerator(test_obj)
                doc += test_doc_section_generator.get_test_section()

        return "\n".join(doc).rstrip() + "\n"

    def _get_test_intro(self, repo_name, repo_url):
        if not self.project.requires_parameters:
            return [
                "Verify that the actual output matches the expected output",
                "(see [Requirements](#requirements)).",
            ]

        doc = [
            f"Every project in the [{repo_name} repo]({repo_url}) should be tested.",
            f"In this section, we specify the set of tests specific to {self.project_title}.",
        ]
        if len(self.project.tests) > 1:
            doc += [
                "In order to keep things simple, we split up the testing as follows:",
                "",
            ]
            doc += [
                "- " + _get_test_section_title(test_obj)
                for test_obj in self.project.tests.values()
            ]

        return doc + [""]


def _get_test_section_title(test_obj):
    return test_obj.name.replace("_", " ").title() + " Tests"


class TestDocSectionGenerator:
    __test__ = False  # Indicate this is not a test

    def __init__(self, test_obj):
        self.test_obj = test_obj
        self.test_obj_name = _get_test_section_title(test_obj)

    def get_test_section(self):
        return (
            self._get_test_section_header()
            + self._get_test_table_header()
            + self._get_test_table()
        )

    def _get_test_section_header(self):
        return [f"### {self.test_obj_name}", ""]

    def _get_test_table_header(self):
        cells = ["Description"] + self.test_obj.inputs
        if self._any_test_output_is_different():
            cells.append("Output")

        return [
            _cells_to_table_line(cells),
            _cells_to_table_line("-" * len(cell) for cell in cells),
        ]

    def _any_test_output_is_different(self):
        if len(self.test_obj.params) < 2:
            return True

        first_expected = self.test_obj.params[0].expected
        return any(
            test_param.expected != first_expected
            for test_param in self.test_obj.params[1:]
        )

    def _get_test_table(self):
        doc = []
        has_output_column = self._any_test_output_is_different()
        num_input_params = len(self.test_obj.inputs)
        for test_param in self.test_obj.params:
            output = test_param.expected
            cells = [test_param.name.title()]
            if test_param.input is None:
                inputs = []
            else:
                inputs = shlex.split(test_param.input)
                extra_inputs = inputs[num_input_params:]
                inputs = inputs[:num_input_params]
                cells += [quote(value) for value in inputs]
                if extra_inputs:
                    cells[-1] += " " + " ".join(quote(value) for value in extra_inputs)

            cells += [""] * (num_input_params - len(inputs))

            if has_output_column:
                if isinstance(output, str):
                    cells.append(quote(output))
                else:
                    cells.append("<br>".join(quote(item) for item in output))

            doc.append(_cells_to_table_line(cells))

        if not has_output_column:
            doc += [
                "",
                "All of these tests should output the following:",
                "",
                "```",
                self.test_obj.params[0].expected,
                "```",
            ]

        return doc + [""]


def _cells_to_table_line(cells):
    return "| " + " | ".join(cells) + " |"



import argparse
import sys

from glotter.source import get_sources
from glotter.settings import Settings
from glotter.utils import error_and_exit
from glotter.download import download, remove_images
from glotter.test import test


def batch(args):
    # Validate arguments
    if args.num_batches < 1:
        error_and_exit("Number of batches must be at least 1")

    if args.batch is not None and (args.batch < 1 or args.batch > args.num_batches):
        error_and_exit(f"Batch number must be from 1 to {args.num_batches}")

    # Get all of the languages
    all_sources = get_sources(Settings().source_root)
    languages = sorted(
        {
            source.language.lower()
            for sources in all_sources.values()
            for source in sources
        }
    )
    num_languages = len(languages)
    num_batches = min(num_languages, args.num_batches)

    # Determine starting and ending batch
    if args.batch is None:
        first_batch = 0
        last_batch = num_batches
    elif args.batch <= num_batches:
        first_batch = args.batch - 1
        last_batch = args.batch
    else:
        sys.exit(0)

    # For each batch
    exit_code = 0
    for n in range(first_batch, last_batch):
        # Get languages for this batch
        batch_args = argparse.Namespace(
            source=None,
            project=None,
            language=set(
                languages[
                    (n * num_languages // num_batches) : (
                        (n + 1) * num_languages // num_batches
                    )
                ]
            ),
            parallel=args.parallel,
        )

        # Download images for this batch
        _display_batch("Downloading images", n, num_batches)
        containers = download(batch_args)

        # Run tests for this batch
        try:
            _display_batch("Testing", n, num_batches)
            test(batch_args)
        except SystemExit as e:
            exit_code = exit_code or int(e.code)

        # If removing images, remove images for this batch
        if args.remove:
            _display_batch("Removing images", n, num_batches)
            remove_images(containers, args.parallel)

    sys.exit(exit_code)


def _display_batch(prefix, n, num_batches):
    print(f"\n*** {prefix} for batch {n + 1} of {num_batches} ***", flush=True)
